package cn.ac.ios.automata.parser.ba;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.OutputStream;
import java.io.PrintStream;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import automata.FAState;
import automata.FiniteAutomaton;
import cn.ac.ios.automata.parser.Parser;
import cn.ac.ios.options.Options;
import gnu.trove.map.TObjectCharMap;
import gnu.trove.map.hash.TObjectCharHashMap;
 
public class BAParser implements Parser {
	
	private Map<String, String> atomCharStrMap ;
	private TObjectCharMap<String> atomStrCharMap ;
	private Map<String, FAState> strStateMap = new HashMap<>();
	
	// this class only allow the characters
	private FiniteAutomaton automaton;
	
	public BAParser(String file) {
		this.atomStrCharMap = new TObjectCharHashMap<>();
		this.atomCharStrMap = new HashMap<>();
		this.automaton = new FiniteAutomaton();
		try {
			FileInputStream inputStream = new FileInputStream(new File(file));
			JBAParser parser = new JBAParser(inputStream);
			parser.parse(this);
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
	}
	
	// share alphabet with other parser
	public BAParser(String file, BAParser otherParser) {
		this.atomCharStrMap = otherParser.atomCharStrMap;
		this.atomStrCharMap = otherParser.atomStrCharMap;
		
		this.automaton = new FiniteAutomaton();
		try {
			FileInputStream inputStream = new FileInputStream(new File(file));
			JBAParser parser = new JBAParser(inputStream);
			parser.parse(this);
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
	}

	@Override
	public FiniteAutomaton parse() {
		return automaton;
	}

	@Override
	public void print(FiniteAutomaton fa, OutputStream out) {
		PrintStream printer = new PrintStream(out);
		if(Options.dot) {
			printer.print("//Buechi \n");
			printer.print("digraph {\n");
	        
	        Set<FAState> states = fa.states;
	        for (FAState state : states) {
	        	printer.print("  " + state.id + " [label=\"" +  state.id + "\"");
	            if(fa.F.contains(state)) printer.print(", shape = doublecircle");
	            else printer.print(", shape = circle");
	            printer.print("];\n");
	            for (String label : fa.getAllTransitionSymbols()) {
	            	Set<FAState> succs = state.getNext(label);
	            	if(succs == null) continue;
	            	for(FAState succ : succs) {
	            		printer.print("  " + state.id + " -> " + succ.id + " [label=\"" + atomCharStrMap.get(label) + "\"];\n");
	            	}
	            }
	        }	
	        printer.print("  " + states.size() + " [label=\"\", shape = plaintext];\n");
	        FAState initState = fa.getInitialState();
	        printer.print("  " + states.size() + " -> " + initState.id + " [label=\"\"];\n");
	        printer.print("}\n\n");
		}else {
			// first output initial state
			printer.print("[" + fa.getInitialState().id + "]\n");
			// transitions
			for(FAState from : fa.states) {
				Iterator<String> nextIt = from.nextIt();
				while(nextIt.hasNext()) {
					String label = nextIt.next();
					for(FAState to : from.getNext(label)) {
						printer.print(atomCharStrMap.get(label) + "," + "[" + from.id + "]->[" + to.id + "]\n");
					}
				}
			}
			
			for(FAState f : fa.F) {
				printer.print("[" + f.id + "]\n");
			}
		}

	}

	@Override
	public void close() {		
	}
	
	protected void setInitial(String state) {
		FAState init = getState(state);
		automaton.setInitialState(init);
	}
	
	protected void addTransition(String source, String target, String ap) {
		FAState st = getState(source);
		FAState tg = getState(target);
		automaton.addTransition(st, tg, getCharFromString(ap) + "");
	}
	
	protected void setAccepting(String state) {
		FAState f = getState(state);
		automaton.F.add(f);
	}
	
	protected void parseBegin() {
		
	}
	
	protected void parseEnd() {
		
		if(automaton.F.isEmpty()) {
			for(FAState f : automaton.states) {
				automaton.F.add(f);
			}
		}
	}
	
	// we reserve '$' sign for L dollar automaton
	private char getCharFromString(String label) {
		char ch = 0;
		if(atomStrCharMap.containsKey(label)) {
			return atomStrCharMap.get(label);
		}
		
		ch = (char) atomStrCharMap.size();
		if(ch >= '$' )   ch ++; // reserve '$' sign
		atomStrCharMap.put(label, ch);
		atomCharStrMap.put(ch + "", label);
		return ch;
	}
	
	private FAState getState(String str) {
		FAState state = strStateMap.get(str);
		if(state == null) {
			state = automaton.createState();
			strStateMap.put(str, state);
		}
		return state;
	}
	
	public String translate(List<String> strs) {
		StringBuilder builder = new StringBuilder();
		for(String str : strs) {
			builder.append(atomCharStrMap.get(str));
		}
		return builder.toString();
	}
	
	public String translate(String strs) {
		StringBuilder builder = new StringBuilder();
		for(int i = 0; i < strs.length(); i ++) {
			builder.append(atomCharStrMap.get("" + strs.charAt(i)));
		}
		return builder.toString();
	}

}
