/* Written by Yong Li, Depeng Liu                                       */
/* Copyright (c) 2016                  	                               */
/* This program is free software: you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* This program is distributed in the hope that it will be useful,      */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License    */
/* along with this program.  If not, see <http://www.gnu.org/licenses/>.*/

package cn.ac.ios.learner.dfa.tree;

import cn.ac.ios.automata.words.WordManager;
import cn.ac.ios.automata.words.Word;

import cn.ac.ios.learner.tree.LearnerTree;
import cn.ac.ios.query.MembershipOracle;
import cn.ac.ios.query.Query;
import cn.ac.ios.query.QuerySimple;
import cn.ac.ios.table.ExprValue;

/**
 * An improved tree-based DFA learning algorithm which uses a suffix of the counterexample
 * to refine the conjecture upon receiving 'no'
 * */

public class LearnerDFATree extends LearnerTree {

	public LearnerDFATree(WordManager contextWord, MembershipOracle<Boolean> membershipOracle) {
		super(contextWord, membershipOracle);
	}

	@Override
	public boolean isOmegaLearning() {
		return false;
	}

	@Override
	public boolean isLeading() {
		return false;
	}

	@Override
	protected boolean processMembershipQuery(Word label, ExprValue valueExpr) {
		Word suffix = valueExpr.get();
		return processMembershipQuery(label, suffix);
	}

	@Override
	protected boolean processMembershipQuery(ExprValue valueExpr) {
		assert false : "illegal call to query";
		return false;
	}

	@Override
	protected boolean processMembershipQuery(Word prefix, Word suffix) {
		Query<Boolean> query = new QuerySimple<>(null, prefix, suffix, -1);
		return membershipOracle.answerMembershipQuery(query);
	}
	
	private boolean isCEAccepting;
	@Override
	protected ExprValue getCounterExampleWord(Query<Boolean> query) {
		if(query.getQueryAnswer() == null) {
			// for direct use of DFA
			isCEAccepting = processMembershipQuery(query.getPrefix(), query.getSuffix());
		}else {
			isCEAccepting = query.getQueryAnswer();
		}
		return getExprValueWord(query.getQueriedWord());
	}

	@Override
	protected CeAnalyzer getCeAnalyzerInstance(ExprValue exprValue) {
		return new CeAnalyzerDFA(exprValue);
	}
	
	private class CeAnalyzerDFA extends CeAnalyzer {

		public CeAnalyzerDFA(ExprValue exprValue) {
			super(exprValue);
		}

		@Override
		public void analyze() {
			this.leafBranch = getHashableValueBoolean(isCEAccepting);
			this.nodePrevBranch = getHashableValueBoolean(!isCEAccepting);
			// only has one leaf
			if(tree.getRoot().isLeaf()) {
				this.wordExpr = getExprValueWord(contextWord.getEmptyWord());
				this.nodePrev = tree.getRoot();
				this.wordLeaf = getExprValueWord(exprValue.get());
				return ;
			}
			
			Word wordCE = this.exprValue.get();
			// get the initial state from automaton
		    int letterNr = 0, stateCurr = -1, statePrev = getInitState();
			// starting state is lambda
			for(; letterNr < wordCE.length(); letterNr ++) {
				stateCurr = computeNextState(statePrev, wordCE.getLetter(letterNr));
				Word prefix = states.get(stateCurr).label;   // S(j)
				Word suffix = wordCE.getSuffix(letterNr + 1); // y[j+1...n]
				boolean resultCurr = processMembershipQuery(prefix, suffix);; // Sj y[j+1..n]
				if(isCEAccepting != resultCurr) {
					break;
				}
				statePrev = stateCurr; // record S(j)
			}
			
			Word wordPrev = states.get(statePrev).label;         // S(j-1)
			this.wordExpr = getExprValueWord(wordCE.getSuffix(letterNr + 1));  // y[j+1..n]
			this.wordLeaf = getExprValueWord(wordPrev.append(wordCE.getLetter(letterNr))); // S(j-1)y[j]
			this.nodePrev = states.get(stateCurr).node;          // S(j)
		}
		
	}

	@Override
	public Word getStateLabel(int state) {
		// TODO Auto-generated method stub
		return null;
	}

}
