package cn.ac.ios.roll.complement;


import java.io.PrintStream;
import java.util.ArrayList;
import java.util.List;

import automata.FiniteAutomaton;
import cn.ac.ios.automata.words.Alphabet;
import cn.ac.ios.automata.words.Word;
import cn.ac.ios.automata.words.WordManager;
import cn.ac.ios.learner.fdfa.LearnerFDFA;
import cn.ac.ios.options.Options;
import cn.ac.ios.oracle.BuechiRunner;
import cn.ac.ios.oracle.teacher.Teacher;
import cn.ac.ios.query.EquivalenceOracle;
import cn.ac.ios.query.Query;
import cn.ac.ios.query.QuerySimple;
import cn.ac.ios.roll.Statistics;
import cn.ac.ios.util.AutomatonPrinter;
import cn.ac.ios.util.BuechiBuilder;
import cn.ac.ios.util.NullSystemOutput;
import cn.ac.ios.util.OutputStreamNull;
import cn.ac.ios.util.PrintStreamNull;
import cn.ac.ios.util.Timer;
import cn.ac.ios.value.Type;
import cn.ac.ios.value.Value;
import cn.ac.ios.value.ValueManager;
import oracle.EqResult;
import oracle.InclusionChecker;
import oracle.IntersectionCheck;

/** 
 * the teacher makes use of the complement of FDFA
 * */
public class ComplementTeacherFDFA implements Teacher<LearnerFDFA, EqResult, Boolean> {

	private FiniteAutomaton B;
	private WordManager wordManager;
	private final InclusionChecker inclusionChecker = new InclusionChecker();
	
	public ComplementTeacherFDFA(WordManager wordManager, FiniteAutomaton B) {
		assert wordManager != null && B != null;
		this.wordManager = wordManager;
		this.B = B;
		if(Options.verbose) this.inclusionChecker.setVerbose();
		// initialize alphabet
		ValueManager contextValue = wordManager.getContextValue();
		Type typeString = contextValue.newTypeLetter(String.class);
		Value valueLetter = typeString.newValue();
		Alphabet alphabet = wordManager.getAlphabet();

		for(String letter : B.alphabet) {
			valueLetter.set(letter);
			alphabet.add(valueLetter.clone());
		}
		Options.log.println("alphabet size: " + alphabet.size());
		alphabet.setImmutable();
		valueLetter.set("");
		wordManager.setLetterSplitter(valueLetter);
	}
	
	@Override
	public Query<EqResult> answerEquivalenceQuery(LearnerFDFA learner) {
		EqResult result = new EqResult();
		Timer timer = new Timer();
		timer.start();
		String prefixStr = "";
		String suffixStr = "";

		Options.log.println("Translating FDFA to Under Buechi automaton ...");
		FiniteAutomaton BF = BuechiBuilder.build(learner, true);
		Statistics.hypothesis = BF;
		++ this.numInterBFandB;
		Options.log.println("Checking the intersection of B(F) (" + BF.states.size() + ") and B ("+ B.states.size() + ")...");
		long t = timer.getCurrentTime();
		IntersectionCheck checker = new IntersectionCheck(BF, B);
		boolean isEmpty = checker.checkEmptiness();
		t = timer.getCurrentTime() - t;
		this.timeInterBFandB += t;
		if(Options.verbose) {
			Options.log.println("Hypothesis for complementation");
			AutomatonPrinter.print(BF, Options.log.getOutputStream());
		}
		
		if(! isEmpty) {
			// we have omega word in FDFA which should not be there
			checker.computePath();
			for(String letter : checker.getPrefix()) {
				prefixStr += letter;
			}

			for(String letter : checker.getSuffix()) {
				suffixStr += letter;
			}
			
			result.isEqual = false;
			result.isCeInTarget = true;
			
		}else {
			// intersection check for A and not A
			FiniteAutomaton BFC = BuechiBuilder.buildNegNFA(learner);
			
			// if optimization is enabled
			if(BuechiC.optimize) {
				++ this.numInterBFandBFC;
				Options.log.println("Checking the intersection of B(F) (" + BF.states.size() + "," + BFC.trans + ") and B(F^c) ("+ BFC.states.size() + "," + BFC.trans + ")...");
				t = timer.getCurrentTime();
				checker = new IntersectionCheck(BFC, BF);
				isEmpty = checker.checkEmptiness();
				t = timer.getCurrentTime() - t;
				this.timeBFandBFC += t;
				if(Options.verbose) {
					Options.log.println("Under approximation of FDFA complementation");
					AutomatonPrinter.print(BFC, Options.log.getOutputStream());
				}
			}else {
				isEmpty = true;
			}

			if(! isEmpty) {
				// we have omega word in FDFA which should not be there
				checker.computePath();
				for(String letter : checker.getPrefix()) {
					prefixStr += letter;
				}

				for(String letter : checker.getSuffix()) {
					suffixStr += letter;
				}
				
				result.isEqual = false;
				result.isCeInTarget = BuechiRunner.isAccepting(B, checker.getPrefix(), checker.getSuffix());
			}else {
				Options.log.println("Checking the inclusion for B(F^c) (" + BFC.states.size() + ") and B ("+ B.states.size() + ")...");
				NullSystemOutput.setNullOutput();
			    // we have to resort to the equivalence check for BFC
				++this.numBFCLessB;
				t = timer.getCurrentTime();
				result.isEqual = inclusionChecker.isIncluded(BFC, B);
				t = timer.getCurrentTime() - t;
				this.timeBFCLessB += t;
				prefixStr = inclusionChecker.getPrefix();
				suffixStr = inclusionChecker.getSuffix();
		    	NullSystemOutput.restoreOutput();
			}
			
		}
		Options.log.println("Done for checking equivalence...");
		Query<EqResult> query = null;
		result.isCeInTarget = ! result.isCeInTarget;
		
		if(result.isEqual) {
			query = new QuerySimple<>(wordManager.getEmptyWord(), wordManager.getEmptyWord());
			query.answerQuery(result);
		}else {

			Word prefix = wordManager.getWordFromString(prefixStr);
			Word suffix = null;
			if(!suffixStr.equals("")) {
				suffix = wordManager.getWordFromString(suffixStr);
			}else {
				suffix = wordManager.getEmptyWord();
			}
			query = new QuerySimple<>(prefix, suffix);
			query.answerQuery(result);
		}
		timer.stop();
		Statistics.timeEquivalenceQuery += timer.getTimeElapsed();
		++ Statistics.numEquivalenceQuery;
		Statistics.timeLastEquivalenceQuery = timer.getTimeElapsed();
		
		if(Options.verbose) Options.log.println("counter example = " + query);
		return query;
	}
	
	@Override
	public Boolean answerMembershipQuery(Query<Boolean> query) {
		List<String> prefix = new ArrayList<>();
		List<String> suffix = new ArrayList<>();
        
		for(int i = 0; i < query.getPrefix().length(); i ++) {
			prefix.add(wordManager.letterToString(query.getPrefix().getLetter(i)));
		}
		for(int i = 0; i < query.getSuffix().length(); i ++) {
			suffix.add(wordManager.letterToString(query.getSuffix().getLetter(i)));
		}
		
        Timer timer = new Timer();
        timer.start();
        
		boolean result;
		
		if(suffix.isEmpty()) {
			result = false;
		}else {
			result = BuechiRunner.isAccepting(B, prefix, suffix);
		}
		
		timer.stop();
		Statistics.timeMembershipQuery += timer.getTimeElapsed();
		++ Statistics.numMembershipQuery; 
		return ! result; // reverse the result for Buechi automaton
	}
	
	private long numInterBFandB = 0;
	private long numInterBFandBFC = 0;
	private long timeInterBFandB = 0;
	private long timeBFandBFC = 0;
	private long numBFCLessB = 0;
	private long timeBFCLessB = 0;
	
	public void print() {
		Options.log.println("#A/\\B = " + numInterBFandB);
		Options.log.println("t#A/\\B = " + timeInterBFandB + " ms");
		Options.log.println("#A/\\!A = " + numInterBFandBFC);
		Options.log.println("t#A/\\!A = " + timeBFandBFC + " ms");
		Options.log.println("#!A<B = " + numBFCLessB );
		Options.log.println("#t!A<B = " + timeBFCLessB + " ms");
	}

}
