package cn.ac.ios.roll.complement;

import java.io.PrintStream;
import java.util.ArrayList;
import java.util.List;

import automata.FAState;
import automata.FiniteAutomaton;
import cn.ac.ios.automata.words.Alphabet;
import cn.ac.ios.automata.words.Word;
import cn.ac.ios.automata.words.WordManager;
import cn.ac.ios.learner.fdfa.LearnerFDFA;
import cn.ac.ios.options.Options;
import cn.ac.ios.oracle.BuechiRunner;
import cn.ac.ios.oracle.teacher.Teacher;
import cn.ac.ios.query.Query;
import cn.ac.ios.query.QuerySimple;
import cn.ac.ios.roll.Statistics;
import cn.ac.ios.util.AutomatonPrinter;
import cn.ac.ios.util.BuechiBuilder;
import cn.ac.ios.util.NullSystemOutput;
import cn.ac.ios.util.OutputStreamNull;
import cn.ac.ios.util.PrintStreamNull;
import cn.ac.ios.util.Timer;
import cn.ac.ios.util.UtilAutomaton;
import cn.ac.ios.value.Type;
import cn.ac.ios.value.Value;
import cn.ac.ios.value.ValueManager;
import oracle.EqResult;
import oracle.InclusionChecker;
import oracle.IntersectionCheck;

/**
 * the teacher makes use of the universality of B(F) and B
 */
public class ComplementTeacherUniversality implements Teacher<LearnerFDFA, EqResult, Boolean> {
	private final InclusionChecker inclusionChecker;
	private final WordManager contextWord;
	private final FiniteAutomaton targetB;

	public ComplementTeacherUniversality(WordManager contextWord, FiniteAutomaton targetAut) {
		assert targetAut != null && contextWord != null;
		this.contextWord = contextWord;
		this.targetB = targetAut;
		this.inclusionChecker = new InclusionChecker();
		if (Options.verbose)
			this.inclusionChecker.setVerbose();
		// initialize alphabet
		ValueManager contextValue = contextWord.getContextValue();
		Type typeString = contextValue.newTypeLetter(String.class);
		Value valueLetter = typeString.newValue();
		Alphabet alphabet = contextWord.getAlphabet();
		if (Options.learnerLDollar) {
			valueLetter.set(WordManager.getStringDollar());
			alphabet.add(valueLetter.clone());
		}
		for (String letter : targetAut.alphabet) {
			valueLetter.set(letter);
			alphabet.add(valueLetter.clone());
		}
		System.out.println("alphabet size: " + alphabet.size());
		alphabet.setImmutable();
		valueLetter.set("");
		contextWord.setLetterSplitter(valueLetter);
	}

	@Override
	public Query<EqResult> answerEquivalenceQuery(LearnerFDFA learner) {
		EqResult result = new EqResult();
		Timer timer = new Timer();
		timer.start();
		String prefixStr = "";
		String suffixStr = "";

		System.out.println("Translating FDFA to Under Buechi automaton ...");
		FiniteAutomaton BF = BuechiBuilder.build(learner, true);
		Statistics.hypothesis = BF;
		++this.numInterAandB;
		System.out.println("Checking the intersection of B(F) (" + BF.states.size() + ") and B ("
				+ targetB.states.size() + ")...");
		long t = timer.getCurrentTime();
		IntersectionCheck checker = new IntersectionCheck(BF, targetB);
		boolean isEmpty = checker.checkEmptiness();
		t = timer.getCurrentTime() - t;
		this.timeInterAandB += t;
		if (Options.verbose) {
			System.out.println("Hypothesis for complementation");
			AutomatonPrinter.print(BF, System.out);
		}

		if (!isEmpty) {
			// we have omega word in FDFA which should not be there
			checker.computePath();
			for (String letter : checker.getPrefix()) {
				prefixStr += letter;
			}

			for (String letter : checker.getSuffix()) {
				suffixStr += letter;
			}

			result.isEqual = false;
			result.isCeInTarget = true;

		} else {
			// intersection check for A and not A
			if(BuechiCU.optimize) {
				++this.numInterBFandBFC;
				FiniteAutomaton BFC = BuechiBuilder.buildNegNFA(learner);
				//
				System.out.println("Checking the intersection of B(F) (" + BF.states.size() + "," + BFC.trans
						+ ") and B(F^c) (" + BFC.states.size() + "," + BFC.trans + ")...");
				t = timer.getCurrentTime();
				checker = new IntersectionCheck(BFC, BF);
				isEmpty = checker.checkEmptiness();
				t = timer.getCurrentTime() - t;
				this.timeInterBFandBFC += t;
				if (Options.verbose) {
					System.out.println("Under approximation of FDFA complementation");
					AutomatonPrinter.print(BFC, System.out);
				}
			}else {
				isEmpty = true;
			}
			
			if (!isEmpty) {
				// we have omega word in FDFA which should not be there
				checker.computePath();
				for (String letter : checker.getPrefix()) {
					prefixStr += letter;
				}

				for (String letter : checker.getSuffix()) {
					suffixStr += letter;
				}

				result.isEqual = false;
				result.isCeInTarget = BuechiRunner.isAccepting(targetB, checker.getPrefix(), checker.getSuffix());
			} else {

				FiniteAutomaton c = new FiniteAutomaton();

				FAState q = c.createState();
				for (String letter : BF.alphabet) {
					c.addTransition(q, q, letter);
				}

				c.setInitialState(q);
				c.F.add(q);

				dk.brics.automaton.Automaton dkA = UtilAutomaton.convertToDkAutomaton(BF);
				dk.brics.automaton.Automaton dkB = UtilAutomaton.convertToDkAutomaton(targetB);

				dkA = dkA.union(dkB);

				FiniteAutomaton ck = UtilAutomaton.convertToRabitAutomaton(dkA);

				System.out.println("Checking the inclusion for (B(F) U B) (" + ck.states.size() + ") and C (Σ^ω) ("
						+ c.states.size() + ")...");
				NullSystemOutput.setNullOutput();
				// we have to resort to the equivalence check for hypothesisNotA
				++this.numUniCheck;
				t = timer.getCurrentTime();
				result.isEqual = inclusionChecker.isIncluded(c, ck);
				t = timer.getCurrentTime() - t;
				this.timeUniCheck += t;
				prefixStr = inclusionChecker.getPrefix();
				suffixStr = inclusionChecker.getSuffix();
				result.isCeInTarget = false;
				NullSystemOutput.restoreOutput();
			}
		}

		System.out.println("Done for checking equivalence...");
		Query<EqResult> query = null;
		result.isCeInTarget = !result.isCeInTarget;

		if (result.isEqual) {
			query = new QuerySimple<>(contextWord.getEmptyWord(), contextWord.getEmptyWord());
			query.answerQuery(result);
		} else {

			Word prefix = contextWord.getWordFromString(prefixStr);
			Word suffix = null;
			if (!suffixStr.equals("")) {
				suffix = contextWord.getWordFromString(suffixStr);
			} else {
				suffix = contextWord.getEmptyWord();
			}
			query = new QuerySimple<>(prefix, suffix);
			query.answerQuery(result);
		}
		timer.stop();
		Statistics.timeEquivalenceQuery += timer.getTimeElapsed();
		++Statistics.numEquivalenceQuery;
		Statistics.timeLastEquivalenceQuery = timer.getTimeElapsed();

		if (Options.verbose)
			System.out.println("counter example = " + query);
		return query;
	}

	@Override
	public Boolean answerMembershipQuery(Query<Boolean> query) {
		List<String> prefix = new ArrayList<>();
		List<String> suffix = new ArrayList<>();

		for (int i = 0; i < query.getPrefix().length(); i++) {
			prefix.add(contextWord.letterToString(query.getPrefix().getLetter(i)));
		}
		for (int i = 0; i < query.getSuffix().length(); i++) {
			suffix.add(contextWord.letterToString(query.getSuffix().getLetter(i)));
		}

		Timer timer = new Timer();
		timer.start();

		boolean result;

		if (suffix.isEmpty()) {
			result = false;
		} else {
			result = BuechiRunner.isAccepting(targetB, prefix, suffix);
		}

		timer.stop();
		Statistics.timeMembershipQuery += timer.getTimeElapsed();
		++Statistics.numMembershipQuery;
		return !result; // reverse the result for Buechi automaton
	}

	private long numInterAandB = 0;
	private long timeInterAandB = 0;
	private long numInterBFandBFC = 0;
	private long timeInterBFandBFC = 0;
	private long numUniCheck = 0;
	private long timeUniCheck = 0;

	public void print() {
		System.out.println("#B(F)/\\B = " + numInterAandB);
		System.out.println("t#B(F)/\\B = " + timeInterAandB + " ms");
		System.out.println("#B(F)/\\B(F^c) = " + numInterBFandBFC);
		System.out.println("t#B(F)/\\B(F^c) = " + timeInterBFandBFC + " ms");
		System.out.println("#A\\/B = " + numUniCheck);
		System.out.println("t#A\\/B = " + timeUniCheck + " ms");
	}
}
