package cn.ac.ios.roll.inclusion;

import automata.FiniteAutomaton;
import cn.ac.ios.automata.parser.ba.BAParser;
import cn.ac.ios.automata.words.Word;
import cn.ac.ios.automata.words.WordManager;
import cn.ac.ios.learner.fdfa.LearnerFDFA;
import cn.ac.ios.options.Options;
import cn.ac.ios.oracle.BuechiRunner;
import cn.ac.ios.query.EquivalenceOracle;
import cn.ac.ios.query.Query;
import cn.ac.ios.query.QuerySimple;
import cn.ac.ios.roll.Statistics;
import cn.ac.ios.util.AutomatonPrinter;
import cn.ac.ios.util.BuechiBuilder;
import cn.ac.ios.util.NullSystemOutput;

import cn.ac.ios.util.Timer;
import oracle.EqResult;
import oracle.InclusionChecker;
import oracle.IntersectionCheck;

public class InclusionEQOracle implements EquivalenceOracle<LearnerFDFA, Query<EqResult>>{

	private FiniteAutomaton A;
	private FiniteAutomaton B;
	private WordManager wordManager;
	private final InclusionChecker inclusionChecker = new InclusionChecker();
	private BAParser parser;
	
	public InclusionEQOracle(WordManager wordManager, FiniteAutomaton A, FiniteAutomaton B, BAParser parser) {
		this.wordManager = wordManager;
		this.A = A;
		this.B = B;
		if(Options.verbose) this.inclusionChecker.setVerbose();
		this.parser = parser;
	}
	
	private int numInterBandBF;
	private long timeInterBandBF;
	
	private int numInterAandBF;
	private long timeInterAandBF;
	
	private int numInterBFCandBF;
	private long timeInterBFCandBF;
	
	private int numBFCLessB;
	private long timeBFCLessB;
	
	@Override
	public Query<EqResult> answerEquivalenceQuery(LearnerFDFA learner) {
		EqResult result = new EqResult();
		Timer timer = new Timer();
		timer.start();
		String prefixStr = "";
		String suffixStr = "";

		System.out.println("Translating FDFA to Under Buechi automaton ...");
		FiniteAutomaton BF = BuechiBuilder.build(learner, true);
		++ this.numInterBandBF;
		System.out.println("Checking the intersection of BF (" + BF.states.size() + ") and B ("+ B.states.size() + ")...");
		long t = timer.getCurrentTime();
		IntersectionCheck checker = new IntersectionCheck(BF, B);
		boolean isEmpty = checker.checkEmptiness();
		t = timer.getCurrentTime() - t;
		this.timeInterBandBF += t;
		if(Options.verbose) {
			System.out.println("Hypothesis for complementation B");
			AutomatonPrinter.print(BF, System.out);
		}
		
		if(! isEmpty) {
			// we have omega word in FDFA which should not be there
			checker.computePath();
			for(String letter : checker.getPrefix()) {
				prefixStr += letter;
			}

			for(String letter : checker.getSuffix()) {
				suffixStr += letter;
			}
			
			result.isEqual = false;
			result.isCeInTarget = true;
			
		}else {
			// intersection check for A and B(F)
			++ this.numInterAandBF;
			
			System.out.println("Checking the intersection of A (" + A.states.size() + "," + A.trans + ") and B(F) ("+ BF.states.size() + "," + BF.trans + ")...");
			t = timer.getCurrentTime();
			checker = new IntersectionCheck(A, BF);
			isEmpty = checker.checkEmptiness();
			t = timer.getCurrentTime() - t;
			this.timeInterAandBF += t;
			
			if(! isEmpty) {
				// we have found counterexample now
				checker.computePath();
				
				result.isEqual = true;
				System.out.println("Inclusion violated... ");
				System.out.println("prefix: " + parser.translate(checker.getPrefix()));
				System.out.println("loop: " + parser.translate(checker.getSuffix()));
				
			}else {
				FiniteAutomaton BFC = BuechiBuilder.buildNegNFA(learner);
				System.out.println("Checking the intersection for B(F) (" + BF.states.size() + ") and B(F^c) ("+ BFC.states.size() + ")...");
				++ this.numInterBFCandBF;
				t = timer.getCurrentTime();
				checker = new IntersectionCheck(BFC, BF);
				isEmpty = checker.checkEmptiness();
				t = timer.getCurrentTime() - t;
				this.timeInterBFCandBF += t;
				
				if(! isEmpty) {
					// we have found counterexample now
					checker.computePath();
					for(String letter : checker.getPrefix()) {
						prefixStr += letter;
					}

					for(String letter : checker.getSuffix()) {
						suffixStr += letter;
					}
					
					result.isEqual = false;
					result.isCeInTarget = BuechiRunner.isAccepting(B, checker.getPrefix(), checker.getSuffix());;
				}else {
					
				    // we have to resort to the equivalence check for hypothesisNotA
					++this.numBFCLessB;
					System.out.println("Checking the inclusion for B(F^c) (" + BFC.states.size() + ") and B ("+ B.states.size() + ")...");
					if(Options.verbose) {
						System.out.println("B(F^c): ");
						BuechiInclusion.parser.print(BFC, System.out);
					}
					t = timer.getCurrentTime();
					NullSystemOutput.setNullOutput();
	
					result.isEqual = inclusionChecker.isIncluded(BFC, B);
					NullSystemOutput.restoreOutput();
					t = timer.getCurrentTime() - t;
					this.timeBFCLessB += t;
					prefixStr = inclusionChecker.getPrefix();
					suffixStr = inclusionChecker.getSuffix();

					if(result.isEqual) {
						System.out.println("Inclusion holds ...");
						result.isEqual = true;
					}else {
						result.isCeInTarget = false;
						// check whether it is in A
						boolean isAStr = BuechiRunner.isAccepting(A, prefixStr, suffixStr);
						
						if(isAStr) {
							System.out.println("Inclusion violated... ");
							System.out.println("prefix: " + parser.translate(prefixStr));
							System.out.println("loop: " + parser.translate(suffixStr));
							result.isEqual = true;
						}
					}
				}
			}
			
		}
		System.out.println("Done for checking equivalence...");
		Query<EqResult> query = null;
		result.isCeInTarget = ! result.isCeInTarget;
		
		if(result.isEqual) {
			query = new QuerySimple<>(wordManager.getEmptyWord(), wordManager.getEmptyWord());
			query.answerQuery(result);
		}else {

			Word prefix = wordManager.getWordFromString(prefixStr);
			Word suffix = null;
			if(!suffixStr.equals("")) {
				suffix = wordManager.getWordFromString(suffixStr);
			}else {
				suffix = wordManager.getEmptyWord();
			}
			query = new QuerySimple<>(prefix, suffix);
			query.answerQuery(result);
		}
		timer.stop();
		Statistics.timeEquivalenceQuery += timer.getTimeElapsed();
		++ Statistics.numEquivalenceQuery;
		Statistics.timeLastEquivalenceQuery = timer.getTimeElapsed();
		
		if(Options.verbose) System.out.println("counter example = " + query);
		return query;
	}
	
	public void print() {
		Options.log.println("#B(F)/\\B = " + numInterBandBF);
		Options.log.println("t#B(F)/\\B = " + timeInterBandBF + " ms");
		Options.log.println("#A/\\B(F) = " + numInterAandBF);
		Options.log.println("t#A/\\B(F) = " + timeInterAandBF + " ms");
		Options.log.println("#B(F^c)/\\B(F) = " + numInterBFCandBF);
		Options.log.println("t#B(F^c)/\\B(F) = " + timeInterBFCandBF + " ms");
		Options.log.println("#!A<B = " + numBFCLessB );
		Options.log.println("#t!A<B = " + timeBFCLessB + " ms");
	}

}
